@@ -1,7 +1,7 @@
 /*
  *  VHDL code generation for statements.
  *
- *  Copyright (C) 2008-2009  Nick Gasson (nick@nickg.me.uk)
+ *  Copyright (C) 2008-2013  Nick Gasson (nick@nickg.me.uk)
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +30,9 @@
 #include <set>
 #include <algorithm>
 
+static void emit_wait_for_0(vhdl_procedural *proc, stmt_container *container,
+                            ivl_statement_t stmt, vhdl_expr *expr);
+
 /*
  * VHDL has no real equivalent of Verilog's $finish task. The
  * current solution is to use `assert false ...' to terminate
@@ -40,10 +43,10 @@
  * An alternative is to use the VHPI interface supported by
  * some VHDL simulators and implement the $finish functionality
  * in C. This function can be enabled with the flag
- * -puse-vhpi-finish=1. 
+ * -puse-vhpi-finish=1.
  */
-static int draw_stask_finish(vhdl_procedural *proc, stmt_container *container,
-                             ivl_statement_t stmt)
+static int draw_stask_finish(vhdl_procedural *, stmt_container *container,
+                             ivl_statement_t)
 {
    const char *use_vhpi = ivl_design_flag(get_vhdl_design(), "use-vhpi-finish");
    if (strcmp(use_vhpi, "1") == 0) {
@@ -51,9 +54,111 @@ static int draw_stask_finish(vhdl_procedural *proc, stmt_container *container,
       container->add_stmt(new vhdl_pcall_stmt("work.Verilog_Support.Finish"));
    }
    else {
-      container->add_stmt(new vhdl_assert_stmt("SIMULATION FINISHED"));
+      container->add_stmt(
+         new vhdl_report_stmt(new vhdl_const_string("SIMULATION FINISHED"),
+                              SEVERITY_FAILURE));
    }
-   
+
+   return 0;
+}
+
+static char parse_octal(const char *p)
+{
+   assert(*p && *(p+1) && *(p+2));
+   assert(isdigit(*p) && isdigit(*(p+1)) && isdigit(*(p+2)));
+
+   return (*p - '0') * 64
+      + (*(p+1) - '0') * 8
+      + (*(p+2) - '0') * 1;
+}
+
+// Generate VHDL report statements for Verilog $display/$write
+static int draw_stask_display(vhdl_procedural *proc,
+                              stmt_container *container,
+                              ivl_statement_t stmt)
+{
+   vhdl_binop_expr *text = new vhdl_binop_expr(VHDL_BINOP_CONCAT,
+                                               vhdl_type::string());
+
+   const int count = ivl_stmt_parm_count(stmt);
+   int i = 0;
+   while (i < count) {
+      // $display may have an empty parameter, in which case
+      // the expression will be null
+      // The behaviour here seems to be to output a space
+      ivl_expr_t net = ivl_stmt_parm(stmt, i++);
+      if (net == NULL) {
+         text->add_expr(new vhdl_const_string(" "));
+         continue;
+      }
+
+      if (ivl_expr_type(net) == IVL_EX_STRING) {
+         ostringstream ss;
+         for (const char *p = ivl_expr_string(net); *p; p++) {
+            if (*p == '\\') {
+               // Octal escape
+               char ch = parse_octal(p+1);
+               if (ch == '\n') {
+                  // Is there a better way of handling newlines?
+                  // Maybe generate another report statement
+               }
+               else
+                  ss << ch;
+               p += 3;
+            }
+            else if (*p == '%' && *(++p) != '%') {
+               // Flush the output string up to this point
+               text->add_expr(new vhdl_const_string(ss.str()));
+               ss.str("");
+
+               // Skip over width for now
+               while (isdigit(*p)) ++p;
+
+               switch (*p) {
+               case 'm':
+                  // TODO: we can get the module name via attributes
+                  cerr << "Warning: no VHDL translation for %m format code"
+                       << endl;
+                  break;
+               default:
+                  {
+                     assert(i < count);
+                     ivl_expr_t netp = ivl_stmt_parm(stmt, i++);
+                     assert(netp);
+
+                     vhdl_expr *base = translate_expr(netp);
+                     if (NULL == base)
+                        return 1;
+
+                     emit_wait_for_0(proc, container, stmt, base);
+
+                     text->add_expr(base->cast(text->get_type()));
+                  }
+               }
+            }
+            else
+               ss << *p;
+         }
+
+         // Emit any non-empty string data left in the buffer
+         if (!ss.str().empty())
+            text->add_expr(new vhdl_const_string(ss.str()));
+      }
+      else {
+         vhdl_expr *base = translate_expr(net);
+         if (NULL == base)
+            return 1;
+
+         emit_wait_for_0(proc, container, stmt, base);
+
+         text->add_expr(base->cast(text->get_type()));
+      }
+   }
+
+   if (count == 0)
+      text->add_expr(new vhdl_const_string(""));
+
+   container->add_stmt(new vhdl_report_stmt(text));
    return 0;
 }
 
@@ -67,9 +172,9 @@ static int draw_stask(vhdl_procedural *proc, stmt_container *container,
    const char *name = ivl_stmt_name(stmt);
 
    if (strcmp(name, "$display") == 0)
-      return draw_stask_display(proc, container, stmt, true);
+      return draw_stask_display(proc, container, stmt);
    else if (strcmp(name, "$write") == 0)
-      return draw_stask_display(proc, container, stmt, false);
+      return draw_stask_display(proc, container, stmt);
    else if (strcmp(name, "$finish") == 0)
       return draw_stask_finish(proc, container, stmt);
    else {
@@ -93,7 +198,7 @@ static int draw_stask(vhdl_procedural *proc, stmt_container *container,
  *
  * If this block has its own scope with local variables then these
  * are added to the process as local variables and the statements
- * are generated as above. 
+ * are generated as above.
  */
 static int draw_block(vhdl_procedural *proc, stmt_container *container,
                       ivl_statement_t stmt, bool is_last)
@@ -111,7 +216,7 @@ static int draw_block(vhdl_procedural *proc, stmt_container *container,
             (new vhdl_var_decl(make_safe_name(sig), type));
       }
    }
-   
+
    int count = ivl_stmt_block_count(stmt);
    for (int i = 0; i < count; i++) {
       ivl_statement_t stmt_i = ivl_stmt_block_stmt(stmt, i);
@@ -124,34 +229,13 @@ static int draw_block(vhdl_procedural *proc, stmt_container *container,
 /*
  * A no-op statement. This corresponds to a `null' statement in VHDL.
  */
-static int draw_noop(vhdl_procedural *proc, stmt_container *container,
-                     ivl_statement_t stmt)
+static int draw_noop(vhdl_procedural *, stmt_container *container,
+                     ivl_statement_t)
 {
    container->add_stmt(new vhdl_null_stmt());
    return 0;
 }
 
-/*
- * The VHDL code generator inserts `wait for 0 ns' after each
- * not-last-in-block blocking assignment.
- * If this is immediately followed by another `wait for ...' then
- * we might as well not emit the first zero-time wait.
- */
-void prune_wait_for_0(stmt_container *container)
-{   
-   vhdl_wait_stmt *wait0;
-   stmt_container::stmt_list_t &stmts = container->get_stmts();
-   while (stmts.size() > 0
-          && (wait0 = dynamic_cast<vhdl_wait_stmt*>(stmts.back()))) {
-      if (wait0->get_type() == VHDL_WAIT_FOR0) {
-         delete wait0;
-         stmts.pop_back();
-      }
-      else
-         break;
-   }
-}
-
 static vhdl_var_ref *make_assign_lhs(ivl_lval_t lval, vhdl_scope *scope)
 {
    ivl_signal_t sig = ivl_lval_sig(lval);
@@ -167,13 +251,13 @@ static vhdl_var_ref *make_assign_lhs(ivl_lval_t lval, vhdl_scope *scope)
    if (e_off) {
       if ((base = translate_expr(e_off)) == NULL)
          return NULL;
-         
+
       vhdl_type integer(VHDL_TYPE_INTEGER);
       base = base->cast(&integer);
    }
-   
+
    unsigned lval_width = ivl_lval_width(lval);
-   
+
    string signame(get_renamed_signal(sig));
    vhdl_decl *decl = scope->get_decl(signame);
    assert(decl);
@@ -197,7 +281,7 @@ static vhdl_var_ref *make_assign_lhs(ivl_lval_t lval, vhdl_scope *scope)
       // ...and use this new variable as the assignment LHS
       decl = shadow_decl;
    }
-   
+
    vhdl_type *ltype = new vhdl_type(*decl->get_type());
    vhdl_var_ref *lval_ref = new vhdl_var_ref(decl->get_name(), ltype);
    if (base) {
@@ -221,7 +305,7 @@ static bool assignment_lvals(ivl_statement_t stmt, vhdl_procedural *proc,
          return false;
 
       lvals.push_back(lhs);
-   }         
+   }
 
    return true;
 }
@@ -239,9 +323,9 @@ assign_for(vhdl_decl::assign_type_t atype, vhdl_var_ref *lhs, vhdl_expr *rhs)
       return new vhdl_assign_stmt(lhs, rhs);
    case vhdl_decl::ASSIGN_NONBLOCK:
       return new vhdl_nbassign_stmt(lhs, rhs);
-   default:
-      assert(false);
-   }   
+   }
+   assert(false);
+   return NULL;
 }
 
 /*
@@ -263,12 +347,68 @@ bool check_valid_assignment(vhdl_decl::assign_type_t atype, vhdl_procedural *pro
       return true;
 }
 
+// Generate a "wait for 0 ns" statement to emulate the behaviour of
+// Verilog blocking assignment using VHDL signals. This is only generated
+// if we read from the target of a blocking assignment in the same
+// process (i.e. it is only generated when required, not for every
+// blocking assignment). An example:
+//
+//   begin
+//     x = 5;
+//     if (x == 2)
+//       y = 7;
+//   end
+//
+// Becomes:
+//
+//   x <= 5;
+//   wait for 0 ns;    -- Required to implement assignment semantics
+//   if x = 2 then
+//     y <= 7;         -- No need for wait here, not read
+//   end if;
+//
+static void emit_wait_for_0(vhdl_procedural *proc,
+                            stmt_container *container,
+                            ivl_statement_t stmt,
+                            vhdl_expr *expr)
+{
+   vhdl_var_set_t read;
+   expr->find_vars(read);
+
+   bool need_wait_for_0 = false;
+   for (vhdl_var_set_t::const_iterator it = read.begin();
+        it != read.end(); ++it) {
+      if (proc->is_blocking_target(*it))
+         need_wait_for_0 = true;
+   }
+
+   stmt_container::stmt_list_t &stmts = container->get_stmts();
+   bool last_was_wait =
+      !stmts.empty() && dynamic_cast<vhdl_wait_stmt*>(stmts.back());
+
+   if (need_wait_for_0 && !last_was_wait) {
+      debug_msg("Generated wait-for-0 for %s:%d",
+                ivl_stmt_file(stmt), ivl_stmt_lineno(stmt));
+
+      vhdl_seq_stmt *wait = new vhdl_wait_stmt(VHDL_WAIT_FOR0);
+
+      ostringstream ss;
+      ss << "Read target of blocking assignment ("
+         << ivl_stmt_file(stmt)
+         << ":" << ivl_stmt_lineno(stmt) << ")";
+      wait->set_comment(ss.str());
+
+      container->add_stmt(wait);
+      proc->added_wait_stmt();
+   }
+}
+
 // Generate an assignment of type T for the Verilog statement stmt.
 // If a statement was generated then `assign_type' will contain the
 // type of assignment that was generated; this should be initialised
 // to some sensible default.
 void make_assignment(vhdl_procedural *proc, stmt_container *container,
-                     ivl_statement_t stmt, bool blocking,
+                     ivl_statement_t stmt, bool emul_blocking,
                      vhdl_decl::assign_type_t& assign_type)
 {
    list<vhdl_var_ref*> lvals;
@@ -280,27 +420,41 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
    if (ivl_expr_type(rval) == IVL_EX_TERNARY) {
       rhs = translate_expr(ivl_expr_oper2(rval));
       rhs2 = translate_expr(ivl_expr_oper3(rval));
+      if (rhs2 == NULL)
+         return;
    }
    else
       rhs = translate_expr(rval);
    if (rhs == NULL)
       return;
 
+   emit_wait_for_0(proc, container, stmt, rhs);
+   if (rhs2)
+      emit_wait_for_0(proc, container, stmt, rhs2);
+
    if (lvals.size() == 1) {
       vhdl_var_ref *lhs = lvals.front();
       rhs = rhs->cast(lhs->get_type());
-     
+
       ivl_expr_t i_delay;
       vhdl_expr *after = NULL;
-      if ((i_delay = ivl_stmt_delay_expr(stmt)) != NULL)
+      if ((i_delay = ivl_stmt_delay_expr(stmt)) != NULL) {
          after = translate_time_expr(i_delay);
+         if (after == NULL)
+            return;
+
+         emit_wait_for_0(proc, container, stmt, after);
+      }
 
       // Find the declaration of the LHS so we know what type
       // of assignment statement to generate (is it a signal,
       // a variable, etc?)
       vhdl_decl *decl = proc->get_scope()->get_decl(lhs->get_name());
       assign_type = decl->assignment_type();
-      
+
+      if (assign_type == vhdl_decl::ASSIGN_NONBLOCK && emul_blocking)
+          proc->add_blocking_target(lhs);
+
       // A small optimisation is to expand ternary RHSs into an
       // if statement (eliminates a function call and produces
       // more idiomatic code)
@@ -308,16 +462,18 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
          rhs2 = rhs2->cast(lhs->get_type());
          vhdl_var_ref *lhs2 =
             make_assign_lhs(ivl_stmt_lval(stmt, 0), proc->get_scope());
-         
+
          vhdl_expr *test = translate_expr(ivl_expr_oper1(rval));
          if (NULL == test)
             return;
 
-         vhdl_if_stmt *vhdif = new vhdl_if_stmt(test);
+         emit_wait_for_0(proc, container, stmt, test);
 
          if (!check_valid_assignment(decl->assignment_type(), proc, stmt))
             return;
 
+         vhdl_if_stmt *vhdif = new vhdl_if_stmt(test);
+
          // True part
          {
             vhdl_abstract_assign_stmt *a =
@@ -344,7 +500,7 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
       // declaration as initialisers. This optimisation is only
       // performed on assignments of constant values to prevent
       // ordering problems.
-      
+
       // This also has another application: If this is an `initial'
       // process and we haven't yet generated a `wait' statement then
       // moving the assignment to the initialization preserves the
@@ -360,7 +516,7 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
           && !decl->has_initial()
           && rhs->constant()
           && decl->get_type()->get_name() != VHDL_TYPE_ARRAY) {
-            
+
          // If this assignment is not in the top-level container
          // it will not be made on all paths through the code
          // This precludes any future extraction of an initialiser
@@ -368,6 +524,7 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
             decl->set_initial(NULL);   // Default initial value
          else {
             decl->set_initial(rhs);
+            proc->get_scope()->hoisted_initialiser(true);
             delete lhs;
             return;
          }
@@ -379,9 +536,8 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
       vhdl_abstract_assign_stmt *a =
          assign_for(decl->assignment_type(), lhs, rhs);
       container->add_stmt(a);
-          
-      if (after != NULL)
-         a->set_after(after);
+
+      a->set_after(after);
    }
    else {
       // Multiple lvals are implemented by first assigning the complete
@@ -396,20 +552,25 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
       proc->get_scope()->add_decl(tmp_decl);
 
       container->add_stmt(new vhdl_assign_stmt(tmp_decl->make_ref(), rhs));
-      
+
       list<vhdl_var_ref*>::iterator it;
       int width_so_far = 0;
       for (it = lvals.begin(); it != lvals.end(); ++it) {
          vhdl_var_ref *tmp_rhs = tmp_decl->make_ref();
-         
+
          int lval_width = (*it)->get_type()->get_width();
          vhdl_expr *slice_base = new vhdl_const_int(width_so_far);
          tmp_rhs->set_slice(slice_base, lval_width - 1);
-         
+
          ivl_expr_t i_delay;
          vhdl_expr *after = NULL;
-         if ((i_delay = ivl_stmt_delay_expr(stmt)) != NULL)
+         if ((i_delay = ivl_stmt_delay_expr(stmt)) != NULL) {
             after = translate_time_expr(i_delay);
+            if (after == NULL)
+               return;
+
+            emit_wait_for_0(proc, container, stmt, after);
+         }
 
          // Find the declaration of the LHS so we know what type
          // of assignment statement to generate (is it a signal,
@@ -419,7 +580,7 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
 
          if (!check_valid_assignment(decl->assignment_type(), proc, stmt))
             return;
-         
+
          vhdl_abstract_assign_stmt *a =
             assign_for(decl->assignment_type(), *it, tmp_rhs);
          if (after)
@@ -428,10 +589,11 @@ void make_assignment(vhdl_procedural *proc, stmt_container *container,
          container->add_stmt(a);
 
          width_so_far += lval_width;
+
+         if (assign_type == vhdl_decl::ASSIGN_NONBLOCK && emul_blocking)
+            proc->add_blocking_target(*it);
       }
    }
-
-   return;
 }
 
 /*
@@ -451,29 +613,13 @@ static int draw_nbassign(vhdl_procedural *proc, stmt_container *container,
 }
 
 static int draw_assign(vhdl_procedural *proc, stmt_container *container,
-                       ivl_statement_t stmt, bool is_last)
+                       ivl_statement_t stmt)
 {
    vhdl_decl::assign_type_t assign_type = vhdl_decl::ASSIGN_NONBLOCK;
-   if (proc->get_scope()->allow_signal_assignment()) {
-      // Blocking assignment is implemented as non-blocking assignment
-      // followed by a zero-time wait
-      // This follows the Verilog semantics fairly closely.
+   bool emulate_blocking = proc->get_scope()->allow_signal_assignment();
 
-      make_assignment(proc, container, stmt, false, assign_type);
+   make_assignment(proc, container, stmt, emulate_blocking, assign_type);
 
-      // Don't generate a zero-wait if either:
-      // a) this is the last statement in the process
-      // c) a blocking assignment was generated
-      if (!is_last && assign_type == vhdl_decl::ASSIGN_NONBLOCK) {
-         prune_wait_for_0(container);
-         container->add_stmt
-            (new vhdl_wait_stmt(VHDL_WAIT_FOR0));
-         proc->added_wait_stmt();
-      }
-   }
-   else
-      make_assignment(proc, container, stmt, true, assign_type);
-   
    return 0;
 }
 
@@ -502,8 +648,6 @@ static int draw_delay(vhdl_procedural *proc, stmt_container *container,
          return 1;
    }
 
-   prune_wait_for_0(container);
-
    ivl_statement_t sub_stmt = ivl_stmt_sub_stmt(stmt);
    vhdl_wait_stmt *wait =
       new vhdl_wait_stmt(VHDL_WAIT_FOR, time);
@@ -511,7 +655,7 @@ static int draw_delay(vhdl_procedural *proc, stmt_container *container,
    // Remember that we needed a wait statement so if this is
    // a process it cannot have a sensitivity list
    proc->added_wait_stmt();
-   
+
    container->add_stmt(wait);
 
    // Expand the sub-statement as well
@@ -519,12 +663,12 @@ static int draw_delay(vhdl_procedural *proc, stmt_container *container,
    // is caught here instead
    if (ivl_statement_type(sub_stmt) != IVL_ST_NOOP)
       draw_stmt(proc, container, sub_stmt);
-   
+
    // Any further assignments occur after simulation time 0
    // so they cannot be used to initialise signal declarations
    // (if this scope is an initial process)
    proc->get_scope()->set_initializing(false);
-   
+
    return 0;
 }
 
@@ -582,19 +726,19 @@ static bool draw_synthesisable_wait(vhdl_process *proc, stmt_container *containe
 {
    // At the moment this only detects FFs with an asynchronous reset
    // All other code will fall back on the default draw_wait
-   
+
    // Store a set of the edge triggered signals
    // The second item is true if this is positive-edge
    set<ivl_nexus_t> edge_triggered;
-   
+
    const int nevents = ivl_stmt_nevent(stmt);
-   
+
    for (int i = 0; i < nevents; i++) {
       ivl_event_t event = ivl_stmt_events(stmt, i);
 
       if (ivl_event_nany(event) > 0)
          return false;
-      
+
       int npos = ivl_event_npos(event);
       for (int j = 0; j < npos; j++)
          edge_triggered.insert(ivl_event_pos(event, j));
@@ -625,6 +769,11 @@ static bool draw_synthesisable_wait(vhdl_process *proc, stmt_container *containe
    set<ivl_nexus_t> test_nexuses;
    get_nexuses_from_expr(ivl_stmt_cond_expr(sub_stmt), test_nexuses);
 
+   // If the test is not a simple function of one variable then this
+   // template will not work
+   if (test_nexuses.size() != 1)
+      return false;
+
    // Now subtracting this set from the set of edge triggered events
    // should leave just one nexus, which is hopefully the clock.
    // If not, then we fall back on the default draw_wait
@@ -661,13 +810,13 @@ static bool draw_synthesisable_wait(vhdl_process *proc, stmt_container *containe
             edge = new vhdl_fcall("falling_edge", vhdl_type::boolean());
    }
    assert(edge);
-   
+
    edge->add_expr(nexus_to_var_ref(proc->get_scope(), *clock_net.begin()));
 
    // Draw the clocked branch
    // For an asynchronous reset we just want this around the else branch,
    stmt_container *else_container = body->add_elsif(edge);
-   
+
    draw_stmt(proc, else_container, ivl_stmt_cond_false(sub_stmt));
 
    if (proc->contains_wait_stmt()) {
@@ -707,7 +856,7 @@ static bool draw_synthesisable_wait(vhdl_process *proc, stmt_container *containe
  * The difficulty stems from VHDL's restriction that a process with
  * a sensitivity list may not contain any `wait' statements: we need
  * to generate these to accurately model some Verilog statements.
- * 
+ *
  * The steps followed are:
  *  1) Determine whether this is the top-level statement in the process
  *  2) If this is top-level, call draw_synthesisable_wait to see if the
@@ -743,17 +892,20 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
 
    // If this container is the top-level statement (i.e. it is the
    // first thing inside a process) then we can extract these
-   // events out into the sensitivity list
-   bool is_top_level = container == proc->get_container()
-      && container->empty();
+   // events out into the sensitivity list as long as we haven't
+   // promoted any preceding assignments to initialisers
+   bool is_top_level =
+      container == proc->get_container()
+      && container->empty()
+      && !proc->get_scope()->hoisted_initialiser();
 
-   // See if this can be implemented in a more idomatic way before we
+   // See if this can be implemented in a more idiomatic way before we
    // fall back on the generic translation
    if (is_top_level && draw_synthesisable_wait(proc, container, stmt))
       return 0;
 
    int nevents = ivl_stmt_nevent(stmt);
-   
+
    bool combinatorial = true;  // True if no negedge/posedge events
    for (int i = 0; i < nevents; i++) {
       ivl_event_t event = ivl_stmt_events(stmt, i);
@@ -770,11 +922,11 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
 
       vhdl_wait_stmt *wait = NULL;
       if (proc->contains_wait_stmt() || !is_top_level)
-         wait = new vhdl_wait_stmt(VHDL_WAIT_ON); 
-      
+         wait = new vhdl_wait_stmt(VHDL_WAIT_ON);
+
       for (int i = 0; i < nevents; i++) {
          ivl_event_t event = ivl_stmt_events(stmt, i);
-         
+
          int nany = ivl_event_nany(event);
          for (int j = 0; j < nany; j++) {
             ivl_nexus_t nexus = ivl_event_any(event, j);
@@ -794,7 +946,7 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
    else {
       // Build a test expression to represent the edge event
       // If this process contains no `wait' statements and this
-      // is the top-level container then we
+      // is the top-level container, then we
       // wrap it in an `if' statement with this test and add the
       // edge triggered signals to the sensitivity, otherwise
       // build a `wait until' statement at the top of the process
@@ -803,22 +955,22 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
 
       stmt_container tmp_container;
       draw_stmt(proc, &tmp_container, ivl_stmt_sub_stmt(stmt), true);
-      
+
       for (int i = 0; i < nevents; i++) {
          ivl_event_t event = ivl_stmt_events(stmt, i);
-         
+
          int nany = ivl_event_nany(event);
          for (int j = 0; j < nany; j++) {
             ivl_nexus_t nexus = ivl_event_any(event, j);
             vhdl_var_ref *ref = nexus_to_var_ref(proc->get_scope(), nexus);
-            
+
             ref->set_name(ref->get_name() + "'Event");
             test->add_expr(ref);
 
             if (!proc->contains_wait_stmt() && is_top_level)
                proc->add_sensitivity(ref->get_name());
          }
-         
+
          int nneg = ivl_event_nneg(event);
          for (int j = 0; j < nneg; j++) {
             ivl_nexus_t nexus = ivl_event_neg(event, j);
@@ -826,13 +978,13 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
             vhdl_fcall *detect =
                new vhdl_fcall("falling_edge", vhdl_type::boolean());
             detect->add_expr(ref);
-            
+
             test->add_expr(detect);
 
             if (!proc->contains_wait_stmt() && is_top_level)
                proc->add_sensitivity(ref->get_name());
          }
-         
+
          int npos = ivl_event_npos(event);
          for (int j = 0; j < npos; j++) {
             ivl_nexus_t nexus = ivl_event_pos(event, j);
@@ -840,14 +992,14 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
             vhdl_fcall *detect =
                new vhdl_fcall("rising_edge", vhdl_type::boolean());
             detect->add_expr(ref);
-         
+
             test->add_expr(detect);
 
             if (!proc->contains_wait_stmt() && is_top_level)
                proc->add_sensitivity(ref->get_name());
          }
       }
-      
+
       if (proc->contains_wait_stmt() || !is_top_level) {
          container->add_stmt(new vhdl_wait_stmt(VHDL_WAIT_UNTIL, test));
          container->move_stmts_from(&tmp_container);
@@ -863,9 +1015,9 @@ static int draw_wait(vhdl_procedural *_proc, stmt_container *container,
 
          container->add_stmt(edge_detect);
       }
-         
+
    }
-   
+
    return 0;
 }
 
@@ -875,8 +1027,11 @@ static int draw_if(vhdl_procedural *proc, stmt_container *container,
    vhdl_expr *test = translate_expr(ivl_stmt_cond_expr(stmt));
    if (NULL == test)
       return 1;
-   
+
+   emit_wait_for_0(proc, container, stmt, test);
+
    vhdl_if_stmt *vhdif = new vhdl_if_stmt(test);
+   container->add_stmt(vhdif);
 
    ivl_statement_t cond_true_stmt = ivl_stmt_cond_true(stmt);
    if (cond_true_stmt)
@@ -886,8 +1041,6 @@ static int draw_if(vhdl_procedural *proc, stmt_container *container,
    if (cond_false_stmt)
       draw_stmt(proc, vhdif->get_else_container(), cond_false_stmt, is_last);
 
-   container->add_stmt(vhdif);
-   
    return 0;
 }
 
@@ -904,7 +1057,7 @@ static vhdl_var_ref *draw_case_test(vhdl_procedural *proc, stmt_container *conta
    if (typeid(*test) != typeid(vhdl_var_ref)) {
       const char *tmp_name = "Verilog_Case_Ex";
       vhdl_type *test_type = new vhdl_type(*test->get_type());
-         
+
       if (!proc->get_scope()->have_declared(tmp_name)) {
          proc->get_scope()->add_decl
             (new vhdl_var_decl(tmp_name, new vhdl_type(*test_type)));
@@ -925,7 +1078,7 @@ static int draw_case(vhdl_procedural *proc, stmt_container *container,
    vhdl_var_ref *test = draw_case_test(proc, container, stmt);
    if (NULL == test)
       return 1;
-   
+
    vhdl_case_stmt *vhdlcase = new vhdl_case_stmt(test);
    container->add_stmt(vhdlcase);
 
@@ -933,7 +1086,7 @@ static int draw_case(vhdl_procedural *proc, stmt_container *container,
    // possible case. So make sure we add an 'others' branch
    // if there isn't a default one.
    bool have_others = false;
-   
+
    int nbranches = ivl_stmt_case_count(stmt);
    for (int i = 0; i < nbranches; i++) {
       vhdl_expr *when;
@@ -947,7 +1100,7 @@ static int draw_case(vhdl_procedural *proc, stmt_container *container,
          when = new vhdl_var_ref("others", NULL);
          have_others = true;
       }
-      
+
       vhdl_case_branch *branch = new vhdl_case_branch(when);
       vhdlcase->add_branch(branch);
 
@@ -960,8 +1113,8 @@ static int draw_case(vhdl_procedural *proc, stmt_container *container,
          new vhdl_case_branch(new vhdl_var_ref("others", NULL));
       others->get_container()->add_stmt(new vhdl_null_stmt());
       vhdlcase->add_branch(others);
-   }      
-   
+   }
+
    return 0;
 }
 
@@ -1027,7 +1180,7 @@ static long get_number_as_long(ivl_expr_t expr)
          }
 
          if (ivl_expr_signed(expr) && bits[nbits-1] == '1' &&
-             nbits < 8*sizeof(long)) imm |= -1L << nbits;
+             nbits < 8*sizeof(long)) imm |= -1UL << nbits;
       }
       break;
    }
@@ -1161,6 +1314,15 @@ static bool process_signal(vhdl_binop_expr *all, vhdl_var_ref *test,
    }
 
    unsigned ewid = ivl_expr_width(expr);
+   if (sizeof(unsigned) >= sizeof(long)) {
+      // Since we will be casting i (constrained by swid) to a long make sure
+      // it will fit into a long. This is actually off by one, but this is the
+      // best we can do since on 32 bit machines an unsigned and long are the
+      // same size.
+      assert(swid <= static_cast<unsigned>(numeric_limits<long>::max()));
+      // We are also going to cast ewid to long so check it as well.
+      assert(ewid <= static_cast<unsigned>(numeric_limits<long>::max()));
+   }
    for (unsigned i = 0; i < swid; i++) {
       // Generate a comparison for this bit position
       vhdl_binop_expr *cmp;
@@ -1170,7 +1332,8 @@ static bool process_signal(vhdl_binop_expr *all, vhdl_var_ref *test,
       // Check if this is an out of bounds access. If this is a casez
       // then check against a constant 'x' for the out of bound bits
       // otherwise skip the check (casex).
-      if (i + sbase >= ewid || i + sbase < 0) {
+      if (static_cast<long>(i) + sbase >= static_cast<long>(ewid) ||
+          static_cast<long>(i) + sbase < 0) {
          if (is_casez) {
             // Get the current test bit.
             type = vhdl_type::nunsigned(width);
@@ -1325,12 +1488,12 @@ int draw_casezx(vhdl_procedural *proc, stmt_container *container,
       return 1;
 
    vhdl_if_stmt *result = NULL;
-   
+
    int nbranches = ivl_stmt_case_count(stmt);
    bool is_casez = ivl_statement_type(stmt) == IVL_ST_CASEZ;
    for (int i = 0; i < nbranches; i++) {
       stmt_container *where = NULL;
-      
+
       ivl_expr_t net = ivl_stmt_case_expr(stmt, i);
       if (net) {
          vhdl_binop_expr *all =
@@ -1342,6 +1505,7 @@ int draw_casezx(vhdl_procedural *proc, stmt_container *container,
                   "expression labels can be translated to VHDL",
                   ivl_stmt_file(stmt), ivl_stmt_lineno(stmt),
                   (is_casez ? "z" : "x"));
+            delete all;
             return 1;
          }
 
@@ -1376,7 +1540,7 @@ int draw_casezx(vhdl_procedural *proc, stmt_container *container,
 
    // We don't actually use the generated `test' expression
    delete test;
-   
+
    return 0;
 }
 
@@ -1392,7 +1556,7 @@ int draw_while(vhdl_procedural *proc, stmt_container *container,
    int rc = draw_stmt(proc, &tmp_container, ivl_stmt_sub_stmt(stmt));
    if (rc != 0)
       return 1;
-   
+
    vhdl_expr *test = translate_expr(ivl_stmt_cond_expr(stmt));
    if (NULL == test)
       return 1;
@@ -1402,9 +1566,13 @@ int draw_while(vhdl_procedural *proc, stmt_container *container,
    vhdl_type boolean(VHDL_TYPE_BOOLEAN);
    test = test->cast(&boolean);
 
+   emit_wait_for_0(proc, container, stmt, test);
+
    vhdl_while_stmt *loop = new vhdl_while_stmt(test);
    draw_stmt(proc, loop->get_container(), ivl_stmt_sub_stmt(stmt));
-   
+
+   emit_wait_for_0(proc, loop->get_container(), stmt, test);
+
    container->add_stmt(loop);
    return 0;
 }
@@ -1436,7 +1604,7 @@ int draw_repeat(vhdl_procedural *proc, stmt_container *container,
    container->add_stmt(loop);
 
    draw_stmt(proc, loop->get_container(), ivl_stmt_sub_stmt(stmt));
-   
+
    return 0;
 }
 
@@ -1452,9 +1620,9 @@ int draw_utask(vhdl_procedural *proc, stmt_container *container,
 
    // TODO: adding some comments to the output would be helpful
 
-   // TOOD: this completely ignores parameters!
+   // TODO: this completely ignores parameters!
    draw_stmt(proc, container, ivl_scope_def(tscope), false);
-   
+
    return 0;
 }
 
@@ -1472,7 +1640,7 @@ int draw_stmt(vhdl_procedural *proc, stmt_container *container,
               ivl_statement_t stmt, bool is_last)
 {
    assert(stmt);
-   
+
    switch (ivl_statement_type(stmt)) {
    case IVL_ST_STASK:
       return draw_stask(proc, container, stmt);
@@ -1481,7 +1649,7 @@ int draw_stmt(vhdl_procedural *proc, stmt_container *container,
    case IVL_ST_NOOP:
       return draw_noop(proc, container, stmt);
    case IVL_ST_ASSIGN:
-      return draw_assign(proc, container, stmt, is_last);
+      return draw_assign(proc, container, stmt);
    case IVL_ST_ASSIGN_NB:
       return draw_nbassign(proc, container, stmt);
    case IVL_ST_DELAY:
@@ -1522,6 +1690,6 @@ int draw_stmt(vhdl_procedural *proc, stmt_container *container,
       error("No VHDL translation for statement at %s:%d (type = %d)",
             ivl_stmt_file(stmt), ivl_stmt_lineno(stmt),
             ivl_statement_type(stmt));
-      return 1;            
+      return 1;
    }
 }
